package org.artifactory.ph;

import org.apache.http.HttpStatus;
import org.artifactory.api.request.DownloadService;
import org.artifactory.api.request.InternalArtifactoryRequest;
import org.artifactory.api.request.JerseyArtifactoryResponse;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.model.PackageDownloadContext;
import org.artifactory.core.api.ph.service.PackageArtifactService;
import org.artifactory.fs.RepoResource;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.request.DownloadRequestContext;
import org.artifactory.resource.ExpiredRepoResource;
import org.artifactory.sapi.fs.VfsFile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.core.Response;
import java.io.InputStream;

import static org.apache.http.HttpStatus.SC_NOT_FOUND;

public class PackageArtifactServiceImpl implements PackageArtifactService {
    private static final Logger log = LoggerFactory.getLogger(PackageArtifactServiceImpl.class);
    private final DownloadService downloadService;
    private final InternalRepositoryService repositoryService;

    public PackageArtifactServiceImpl(DownloadService downloadService, InternalRepositoryService repositoryService) {
        this.downloadService = downloadService;
        this.repositoryService = repositoryService;
    }

    @Override
    public Response download(PackageDownloadContext downloadContext) throws PackageException {
        try {
            JerseyArtifactoryResponse response = new JerseyArtifactoryResponse();
            RepoPath repoPath = RepoPathFactory.create(downloadContext.getRepoKey(), downloadContext.getPath());
            InternalArtifactoryRequest request = new InternalArtifactoryRequest(repoPath, false);
            downloadService.process(request, response);
            return response.build();
        } catch (Exception e) {
            String message = String.format("Failed to process download for path:%s in repo:%s",
                    downloadContext.getPath(), downloadContext.getRepoKey());
            log.error(message);
            throw new PackageException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Internal error", e.getMessage());
        }
    }

    @Override
    public PackageArtifact getArtifact(String repoKey, String path) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(repoKey);
        if (repo == null) {
            throw repoNotFoundException(repoKey, path);
        }
        VfsFile<?> vfsFile = ArtifactExtractor.locateFile(repoKey, path, repositoryService);
        PackageType packageType = PackageType.valueOf(repo.getDescriptor().getType().name().toUpperCase());
        return PackageArtifactMapper.artifact(vfsFile, packageType);

    }

    @Override
    public InputStream getStream(String repoKey, String path) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(repoKey);
        if (repo == null) {
            throw repoNotFoundException(repoKey, path);
        }
        VfsFile<?> vfsFile = ArtifactExtractor.locateFile(repoKey, path, repositoryService);
        return vfsFile.getStream();
    }

    @Override
    public BaseResource getResource(String repoKey, String path) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(repoKey);
        if (repo == null || !repo.isReal()) {
            throw repoNotFoundException(repoKey, path);
        }
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        InternalArtifactoryRequest request = new InternalArtifactoryRequest(repoPath, false);
        return repo.getInfo(new DownloadRequestContext(request));
    }

    @Override
    public BaseResource getResourceFromVirtualCache(String repoKey, String path) throws PackageException {
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        VirtualRepo virtualRepo = repositoryService.virtualRepositoryByKey(repoKey);
        if (virtualRepo == null) {
            throw repoNotFoundException(repoKey, path);
        }
        InternalArtifactoryRequest request = new InternalArtifactoryRequest(repoPath, false);
        RepoResource cachedResource = virtualRepo.getInfoFromVirtualCache(new DownloadRequestContext(request));
        if (cachedResource.isFound() && virtualRepo.isCachedResourceExpired(cachedResource)) {
            cachedResource = new ExpiredRepoResource(cachedResource);
        }
        return cachedResource;
    }

    private static PackageException repoNotFoundException(String repoKey, String path) {
        String message = String.format("Failed to find repository in repo:%s, path:%s", repoKey, path);
        return new PackageException(SC_NOT_FOUND, "Not found", message);
    }

}
