package org.artifactory.build;

import org.artifactory.api.rest.build.ContinueBuildFilter;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.storage.db.build.entity.BuildIdEntity;
import org.artifactory.storage.db.build.service.BuildIdImpl;
import org.glassfish.jersey.internal.guava.Predicates;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;

/**
 * @author Omri Ziv
 */

@Test
public class BuildServiceUtilsTest {

    private final long now = System.currentTimeMillis();

    @Mock
    ArtifactoryHome artifactoryHome;
    @Mock
    ArtifactorySystemProperties artifactorySystemProperties;

    private AutoCloseable autoCloseable;

    @BeforeMethod
    public void setup() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getLongProperty(any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    return Long.parseLong(constantValue.getDefValue().trim());
                });
    }

    @AfterMethod
    public void cleanup() throws Exception {
        ArtifactoryHome.unbind();
        autoCloseable.close();
    }

    public void testIterateFetchFromDb() {
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(10L);
        List<BuildIdEntity> list = BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, Predicates.alwaysTrue(),
                buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                        buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(), buildIdEntity.getBuildRepo()),
                continueBuildFilter);
        assertEquals(list.size(), 10);

    }

    public void testIterateFetchFromDb0Limit() {
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(0L);
        List<BuildIdEntity> list = BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, Predicates.alwaysTrue(),
                buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                        buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(), buildIdEntity.getBuildRepo()),
                continueBuildFilter);
        assertEquals(list.size(), 0);
    }

    public void testIterateFetchFromDbNoResults() {
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(0L);
        List<BuildIdEntity> list =
                BuildServiceUtils.iterateFetchFromDb(x -> Collections.emptyList(), Predicates.alwaysTrue(),
                        buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                                buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(),
                                buildIdEntity.getBuildRepo()),
                        continueBuildFilter);
        assertEquals(list.size(), 0);
    }


    public void testIterateFetchFromDbFilterOdds() {
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(10L);
        List<BuildIdEntity> list =
                BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, build -> build.getBuildId() % 2 == 0,
                        buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                                buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(),
                                buildIdEntity.getBuildRepo()),
                        continueBuildFilter);
        assertEquals(list.size(), 10);
        assertEquals(list.get(0).getBuildId(), 2);
        assertEquals(list.get(9).getBuildId(), 20);

    }

    public void testIterateFetchFromDbFilterOddsStartFrom() {
        long now = System.currentTimeMillis();
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(10L);
        BuildId continueBuildId = new BuildIdImpl(16, "build16", "16", now, BUILDINFO_REPO_KEY);
        continueBuildFilter.setContinueBuildId(continueBuildId);
        List<BuildIdEntity> list =
                BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, build -> build.getBuildId() % 2 == 0,
                        buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                                buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(),
                                buildIdEntity.getBuildRepo()),
                        continueBuildFilter);
        assertEquals(list.size(), 10);
        assertEquals(list.get(0).getBuildId(), 18);
        assertEquals(list.get(9).getBuildId(), 36);

    }

    public void test() {
        BuildCoordinate coordinate =
                BuildServiceUtils.extractCoordinatesFromPath("yevdo-build/2-1592480449869.json",
                        BUILDINFO_REPO_KEY);
        assertNotNull(coordinate);
        assertEquals(coordinate.getBuildName(), "yevdo-build");
        assertEquals(coordinate.getBuildNumber(), "2");
        assertEquals(coordinate.getDate(), 1592480449869L);

        coordinate = BuildServiceUtils
                .extractCoordinatesFromPath("artifactory-build-info:blabla.json", BUILDINFO_REPO_KEY);
        assertNull(coordinate);
    }

    private List<BuildIdEntity> fetchFunction1(ContinueBuildFilter continueBuildFilter) {
        return IntStream.rangeClosed(1, 100)
                .mapToObj(number -> new BuildIdEntity(number, "build-" + number, number + "", now,
                        BUILDINFO_REPO_KEY))
                .filter(buildIdEntity -> continueBuildFilter.getContinueBuildId() == null ||
                        Integer.parseInt(buildIdEntity.getBuildNumber()) >
                                Integer.parseInt(continueBuildFilter.getContinueBuildId().getNumber()))
                .limit(continueBuildFilter.getLimit())
                .collect(Collectors.toList());
    }

}
