package org.artifactory.storage.db.event.service;

import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.event.dao.EventsDao;
import org.artifactory.storage.db.event.dao.NodeEventsCursorDao;
import org.artifactory.storage.db.event.entity.DbNodeEventCursor;
import org.artifactory.storage.db.event.entity.enums.NodeEventCursorStatus;
import org.artifactory.storage.db.event.entity.enums.NodeEventCursorType;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.jfrog.common.ClockUtils;
import org.mockito.ArgumentCaptor;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.List;
import java.util.Optional;

import static org.artifactory.storage.db.event.entity.enums.NodeEventCursorType.getEventIdLogTypes;
import static org.artifactory.storage.db.event.entity.enums.NodeEventCursorType.getTimestampLogTypes;
import static org.fest.assertions.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

/**
 * @author Uriah Levy
 */
@Test
public class NodeEventCursorServiceImplTest {

    private static final String METADATA_OPERATOR_ID = "metadata-dispatcher";
    private static final String FEDERATED_OPERATOR_ID =
            "repo-federated_replication_url_to_repo-cursor";

    private NodeEventCursorService service;
    private NodeEventsCursorDao nodeEventsCursorDao;
    private EventsDao eventsDao;

    @BeforeMethod
    public void setup() {
        eventsDao = mock(EventsDao.class);
        nodeEventsCursorDao = mock(NodeEventsCursorDao.class);
        NodeEventCursorServiceImpl nodeEventCursorService = new NodeEventCursorServiceImpl();
        nodeEventCursorService.setNodeEventsCursorDao(nodeEventsCursorDao);
        nodeEventCursorService.setEventsDao(eventsDao);
        service = nodeEventCursorService;
    }

    public void updateCursor() throws SQLException {
        long eventTimestamp = ClockUtils.epochMillis();
        service.updateOrInsertCursor(
                new NodeEventCursor(METADATA_OPERATOR_ID, eventTimestamp, NodeEventCursorType.METADATA_MIGRATION));
        DbNodeEventCursor eventCursor = DbNodeEventCursor.builder()
                .eventMarker(eventTimestamp)
                .operatorId(METADATA_OPERATOR_ID)
                .type(NodeEventCursorType.METADATA_MIGRATION)
                .build();
        verify(nodeEventsCursorDao).update(eventCursor);
    }

    public void cursorForOperatorId() throws SQLException {
        long eventTimestamp = ClockUtils.epochMillis();
        DbNodeEventCursor dbNodeEventCursor = DbNodeEventCursor.builder()
                .eventMarker(eventTimestamp)
                .operatorId(METADATA_OPERATOR_ID)
                .type(NodeEventCursorType.METADATA_MIGRATION)
                .build();
        when(nodeEventsCursorDao.cursorForOperator(METADATA_OPERATOR_ID)).thenReturn(Optional.of(dbNodeEventCursor));
        assertEquals(service.cursorForOperator(METADATA_OPERATOR_ID),
                new NodeEventCursor(METADATA_OPERATOR_ID, eventTimestamp, NodeEventCursorType.METADATA_MIGRATION));
    }

    @Test(dataProvider = "provideTimestamps")
    public void oldestTimestampForEventLog(long oldestTimestamp, long oldestEventId, long eventIdTimestamp,
            long expected) throws SQLException {
        when(nodeEventsCursorDao.oldestMarkerForTypes(getTimestampLogTypes())).thenReturn(oldestTimestamp);
        when(nodeEventsCursorDao.oldestMarkerForTypes(getEventIdLogTypes())).thenReturn(oldestEventId);
        when(eventsDao.getOldestTimestampByEventIdInclusive(oldestEventId)).thenReturn(eventIdTimestamp);
        assertThat(service.oldestTimestampForEventLogOperator()).isEqualTo(expected);
    }

    @Test
    public void testGetAllCursorsByStatus() throws SQLException {
        when(nodeEventsCursorDao.cursorForStatus(NodeEventCursorStatus.SYNC,
                NodeEventCursorType.FEDERATED_REPLICATION))
                .thenReturn(getDbNodeEvents());
        assertEquals(service.getAllCursorsByStatus(NodeEventCursorStatus.SYNC,
                NodeEventCursorType.FEDERATED_REPLICATION)
                        .size(),
                2);
    }

    @Test
    public void testFailOfGetAllCursorByStatus() throws SQLException {
        try {
            when(nodeEventsCursorDao.cursorForStatus(NodeEventCursorStatus.OUT_OF_SYNC,
                    NodeEventCursorType.FEDERATED_REPLICATION))
                    .thenThrow(new SQLException("not good"));
            service.getAllCursorsByStatus(NodeEventCursorStatus.OUT_OF_SYNC,
                    NodeEventCursorType.FEDERATED_REPLICATION);
        } catch (StorageException se) {
            assertTrue(se.getMessage().contains("OUT_OF_SYNC"));
        }
    }

    @Test
    public void testGetAllOldCursorsByTypeAndStatus() throws SQLException {

        when(nodeEventsCursorDao.getCursorsBy(NodeEventCursorType.getEventIdLogTypes(),
                NodeEventCursorStatus.SYNC, 1639935650L)).thenReturn(getDbNodeEvents());
        assertEquals(service.getEventIdCursorsBy(
                NodeEventCursorStatus.SYNC, 1639935650L)
                        .size(),
                2);
    }

    @Test
    public void testUpdateCursorWithStatus() throws SQLException {
        service.updateCursorWithStatus("opId", NodeEventCursorStatus.OUT_OF_SYNC);
        verify(nodeEventsCursorDao, times(1)).updateCursorsStatus("opId",
                NodeEventCursorStatus.OUT_OF_SYNC);
    }

    @Test
    public void testUpdateWithTimestamp() throws SQLException {
        long lasUpdate = ClockUtils.epochMillis() - 1;
        ArgumentCaptor<Long> argument = ArgumentCaptor.forClass(Long.class);
        service.updateCursorMarker("op", 1L, true);
        verify(nodeEventsCursorDao, times(1))
                .updateWithMarkerAndTimestamp(anyString(), anyLong(), argument.capture());
        assertTrue(argument.getValue() > lasUpdate);
    }

    @Test
    public void testUpdateWithoutTimestamp() throws SQLException {
        service.updateCursorMarker("op", 1L, false);
        verify(nodeEventsCursorDao, times(1)).update("op", 1L);
    }

    @DataProvider
    private Object[][] provideTimestamps() {
        return new Object[][] {
                {100000L, 100, 50000L, 50000L},
                {100000L, 100, 200000L, 100000L},
                {0, 100, 200000L, 200000L},
                {100000L, 0, 0, 100000L},
                {0, 0, 0, 0},
        };
    }

    private List<DbNodeEventCursor> getDbNodeEvents() {
        return List.of(
                DbNodeEventCursor.builder()
                        .eventMarker(1L)
                        .operatorId(FEDERATED_OPERATOR_ID)
                        .type(NodeEventCursorType.FEDERATED_REPLICATION)
                        .status(NodeEventCursorStatus.SYNC)
                        .lastUpdated(1639903154L)
                        .build(),
                DbNodeEventCursor.builder()
                        .eventMarker(2L)
                        .operatorId(FEDERATED_OPERATOR_ID + "2")
                        .type(NodeEventCursorType.FEDERATED_REPLICATION)
                        .status(NodeEventCursorStatus.SYNC)
                        .lastUpdated(1639903155L)
                        .build()
        );
    }
}