package org.artifactory.storage.db.remotebintask.dao;

import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.remotebintask.RemoteBinaryTaskDao;
import org.artifactory.storage.db.remotebintask.entity.RemoteBinaryTask;
import org.artifactory.storage.db.util.JdbcHelper;
import org.jfrog.storage.DbType;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.artifactory.common.ConstantValues.haHeartbeatStaleIntervalSecs;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

public class RemoteBinaryTaskDaoTest {

    @Mock
    private JdbcHelper jdbcHelper;
    @Mock
    private DbService dbService;
    @Mock
    private ArtifactoryDbProperties dbProperties;
    @Mock
    private ResultSet resultSet;
    @Mock
    private ArtifactoryHome home;
    @Mock
    private ArtifactorySystemProperties sysProps;

    private RemoteBinaryTaskDao remoteBinaryTaskDao;

    @BeforeMethod
    private void setup() throws Exception {
        openMocks(this);

        when(sysProps.getLongProperty(eq(ConstantValues.haHeartbeatStaleIntervalSecs))).thenReturn(30 * 1000L);

        ArtifactoryHome.bind(home);

        when(dbProperties.getDbType()).thenReturn(DbType.POSTGRESQL);
        when(dbService.nextId()).thenReturn(1L);
        when(jdbcHelper.executeSelect(anyString())).thenReturn(resultSet);
        when(jdbcHelper.executeSelect(anyString(), any())).thenReturn(resultSet);
        when(resultSet.next()).thenReturn(false);
        when(home.getArtifactoryProperties()).thenReturn(sysProps);
        remoteBinaryTaskDao = new RemoteBinaryTaskDao(jdbcHelper, dbProperties);
    }

    @AfterMethod
    private void teardown() {
        ArtifactoryHome.unbind();
    }

    @DataProvider
    public Object[][] testAddArgsDataProvider() {
        return new Object[][] {
                {null, "localRepoKey"},
                {"sha1", null}
        };
    }

    @Test(dataProvider = "testAddArgsDataProvider", expectedExceptions = IllegalArgumentException.class)
    public void testAddArgs(String sha1, String localRepoKey) throws SQLException {
        remoteBinaryTaskDao.add(1L, sha1, localRepoKey, null);
    }

    @Test
    public void testAdd() throws SQLException {
        long time = System.currentTimeMillis();

        remoteBinaryTaskDao.add(1L,"sha1", "localRepoKey", "remoteUrl");
        ArgumentCaptor<Object> params = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeUpdate(anyString(), params.capture());

        Assert.assertEquals(params.getAllValues().get(0), 1L);
        Assert.assertEquals(params.getAllValues().get(1), "sha1");
        Assert.assertTrue((Long) params.getAllValues().get(2) >= time);
        Assert.assertEquals(params.getAllValues().get(3), "localRepoKey");
        Assert.assertEquals(params.getAllValues().get(4), "remoteUrl".hashCode());
        Assert.assertNull(params.getAllValues().get(5));
        Assert.assertEquals(params.getAllValues().get(6), 0);
        Assert.assertNull(params.getAllValues().get(7));
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void testAllocateArgs() throws SQLException {
        remoteBinaryTaskDao.allocate(1, null);
    }

    @Test
    public void testAllocate() throws SQLException {
        long time = System.currentTimeMillis();
        remoteBinaryTaskDao.allocate(1, "serverId");
        ArgumentCaptor<Object> params = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeUpdate(anyString(), params.capture());

        Assert.assertTrue((Long) params.getAllValues().get(0) >= time);
        Assert.assertEquals(params.getAllValues().get(1), "serverId");
        Assert.assertEquals(params.getAllValues().get(2), 1L);
    }

    @Test
    public void testDeallocateById() throws SQLException {
        remoteBinaryTaskDao.deAllocateById(1);
        ArgumentCaptor<Object> params = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeUpdate(anyString(), params.capture());

        Assert.assertEquals(params.getAllValues().get(0), 1L);
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void testDeallocateByServerIdArgs() throws SQLException {
        remoteBinaryTaskDao.deAllocateByServerId(null);
    }

    @Test
    public void testDeallocateByServerId() throws SQLException {
        remoteBinaryTaskDao.deAllocateByServerId("serverId");
        ArgumentCaptor<Object> params = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeUpdate(anyString(), params.capture());

        Assert.assertEquals(params.getAllValues().get(0), "serverId");
    }

    @Test
    public void testRemove() throws SQLException {
        remoteBinaryTaskDao.remove(1);
        ArgumentCaptor<Object> params = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeUpdate(anyString(), params.capture());

        Assert.assertEquals(params.getAllValues().get(0), 1L);
    }

    @Test
    public void testGetAll() throws SQLException {
        Assert.assertNotNull(remoteBinaryTaskDao.getAll());
        verify(jdbcHelper, times(1)).executeSelect(anyString());
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void testNextArgs() throws SQLException {
        remoteBinaryTaskDao.getTasks(null, 1, 1, 15000, 600);
    }

    @Test
    public void testNoNext() throws SQLException {
        List<RemoteBinaryTask> task = remoteBinaryTaskDao.getTasks("serverId",
                1, 1, 15000, 600);
        Assert.assertEquals(task.size(), 0);
        ArgumentCaptor<Object> params = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeSelect(anyString(), params.capture());
        verify(jdbcHelper, never()).executeUpdate(anyString(), any());
    }

    @Test
    public void testNext() throws SQLException {
        long time = System.currentTimeMillis();
        when(resultSet.next()).thenReturn(true, false);

        when(resultSet.getLong(eq("id"))).thenReturn(1L);
        when(resultSet.getString(eq("sha1"))).thenReturn("sha1");
        when(resultSet.getLong(eq("time"))).thenReturn(System.currentTimeMillis());
        when(resultSet.getString(eq("local_repo_key"))).thenReturn("repo1");
        when(resultSet.getInt(eq("count"))).thenReturn(0);

        RemoteBinaryTask task = remoteBinaryTaskDao.getTasks("server1", 1, 1,
                15000, 600).get(0);
        Assert.assertNotNull(task);

        ArgumentCaptor<Object> selectParams = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeSelect(anyString(), selectParams.capture());
        List<Object> allValues = selectParams.getAllValues();
        Assert.assertEquals((int) (Integer) allValues.get(0), 1);
        Assert.assertEquals((long) allValues.get(1), 15000);
        Assert.assertTrue((long) allValues.get(2) >= time);
        Assert.assertEquals((long) allValues.get(3), 600);
        Assert.assertTrue((Long) allValues.get(4) >= time);
        Assert.assertEquals(SECONDS.toMillis(haHeartbeatStaleIntervalSecs.getInt()), (long) allValues.get(5));
        Assert.assertTrue((Long) allValues.get(6) >= time);


        ArgumentCaptor<Object> updateParams = ArgumentCaptor.forClass(String[].class);
        verify(jdbcHelper, times(1)).executeUpdate(anyString(), updateParams.capture());
        Assert.assertTrue((Long) updateParams.getAllValues().get(0) >= time);
        Assert.assertEquals(updateParams.getAllValues().get(1), "server1");
        Assert.assertEquals(updateParams.getAllValues().get(2), task.getId());
    }
}
